/* ========================================================================

 iops.c  Index Of Patchiness for Stations
 ----------------------------------------

 As the name implies, this programs calculates the index of patchiness for 
 stations. IOPS is run from the commandline and takes an input filename as 
 an arguement.   eg:    ./iops sample_data1.txt

 The input file format is defined as a simple text file of integers on
 separate lines. The first integer is the number of items in the dataset,
 and the remaining integers being the dataset.

 The observed distance is the distance the observed counts have to move to 
 give a regular pattern. The index of pacthiness is the ratio of the 
 observed distance to the average of the distance from the permited counts. 
 
 The probability is the proportion of index-values from randomly arranged 
 counts that were at least as large as the value of the observed index.
 This is used to estimate if the observed arrangement of counts is 
 significantly different from a random arrangement.

 Copyright 2005 
   Biomathematics Research Centre,
   Mathematics and Statistics
   University of Canterbury,
   Private Bag 4800
   CHRISTCHURCH 1
   New Zealand

  Credits: 
  --------
  IOPS Algorithm:  Dr. Jennifer Brown
  Programming:     Mr. Dietrich Radel
          
 ========================================================================*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>

#define MAX_ARRAY_SIZE 1000
#define MAX_DATA 1000
#define MIN_DATA 0
 
float read_datafile(char *filename);
float algorithm    (float *TEMP, float temp_mean);
int   shuffle_array(float *ORIGINAL, float *NEW);

float ARRAY [MAX_ARRAY_SIZE];      /* an array of integers between 0 and 3 */
float STORE [MAX_ARRAY_SIZE];       /* original array values for reference */
int num_stations;            /* an integer storing the size of the array A */
int seed;                 /* Global value to help random number generation */




/* ========================================================================
   Here is where everything starts. Various functions are called from here. 
   ========================================================================*/
int main(int argc, char** argv){

  float observed_mean = 0;     /* stores the the mean of data in the array */
  float observed_distance = 0;          /* obeserved distance of the array */
  float shuffle_distance = 0;           /* obeserved distance of the array */
  float accumulated_distance = 0;
  float average_shuffle_distance = 0;
  float probability, index = 0;
  int i, j, l = 0;
  int count = 1;

  /* Check the commandline. Needs to supply a filename for datafile. */
  if (argc !=2){
    fprintf(stderr, "Usage: algorithm <datafile>\n");
    exit(1);
  }
  
 /* Initialize arrays */
  for (i = 0; i < MAX_ARRAY_SIZE; i++){
    ARRAY[i]=-1; 
    STORE[i]=-1;
  }

  /* Reads data into ARRAY[] from file */
  observed_mean =  read_datafile(argv[1]);    

  /* Backup original ARRAY to refer to when creating permutations */
  for (i = 0; i < MAX_ARRAY_SIZE; i++){
    STORE[i]= ARRAY[i];
  }

  /* Call main algorithm to compute */
  observed_distance = algorithm(&ARRAY[0], observed_mean);    

  for (j = 1; j <= 999; j++){ 
    shuffle_array(&STORE[0], &ARRAY[0]);
    shuffle_distance = algorithm(&ARRAY[0], observed_mean); 
    accumulated_distance = accumulated_distance + shuffle_distance;
    if (shuffle_distance >= observed_distance){
        count++;
    }
  }
  
  
  
  average_shuffle_distance = accumulated_distance / 999;
  index = observed_distance / average_shuffle_distance;
  probability = (float)count / 1000;

  /* Output our data to stdout */
  printf("---------------------------------\n");
  printf("Index Of Patchiness for Stations:\n");
  printf("---------------------------------\n\n");
  printf("Original Dataset:\n");
  for(l=0; l < num_stations; l++){
    printf("%d ",(int) STORE[l]);
  } 
  printf("\n\n");
  printf("Number of stations   = %d\n",     num_stations);
  printf("Observed mean        = %.1f\n",   observed_mean);
  printf("Observed distance    = %.1f\n",   observed_distance);  
  printf("Avg shuffle distance = %.1f\n",   average_shuffle_distance);
  printf("Index of patchiness  = %.3f\n",   index);
  printf("Probability          = %.3f\n\n", probability);
  return 0;
}


/* =====================================================================
   Function to read data from a data file.  The filename is supplied via
   the commandline.
   ===================================================================== */
float read_datafile(char *filename){
  float mean = 0;    /* an integer storing the the mean of data in the array */
  int counter = 0;                                   /* Counter variable.*/
  float temp = 0;
  FILE *fp;                                           /* The input file. */

  /* Standard C file opening routine. */
  if((fp = fopen(filename, "r")) == NULL){
    fprintf(stderr, "can't open data file\n");
    exit(1);
  } 
  
  /* Input number of array elements so we know how many fscanfs to do.   */
  fscanf(fp, "%d", &num_stations);
  
  if (num_stations > MAX_ARRAY_SIZE){
    fprintf(stderr,"Sorry, too much data for the program!\n");
    exit(1);
  }

  /* Read all the data into the array. */
  for(counter=0; counter < num_stations; counter++){
    fscanf(fp, "%f", &ARRAY[counter]);
    if (ARRAY[counter] < MIN_DATA || ARRAY[counter] > MAX_DATA){
      fprintf(stderr,"Sorry, data %.1f is not in the range %d - %d.\n", 
	     ARRAY[counter], MIN_DATA, MAX_DATA);
      exit(1);
    }
    temp = temp+ARRAY[counter];
  }
  mean = temp / num_stations;                            /* Calculate Average. */
  return mean;
}


/* =====================================================================
   This is the ALGORITHM. This function receives a pointer to an array
   of data, calculates the new positions, returns the mean.
   This function modifies any array of floats passed to it.
   ===================================================================== */
float algorithm(float *TEMP, float temp_mean){
  float distance=0;
  float running_total = 0;
  int factor, a, b = 0;

  /* Step through ARRAY from Cell 1 to N) */
  for (a = 0; a < num_stations; a++){
 
    /* put cell value into running total */
    if (TEMP[a] > temp_mean){
      running_total = TEMP[a] - temp_mean; 
      TEMP[a] = temp_mean;
      /* Keep going until RUNNING_TOTAL = 0 */
      if (running_total > 0) {    
	/*Step through TEMP from Cell 0 to N */   
	for (b = 0; b < num_stations; b++) {  
	  if ( TEMP[b] < temp_mean ) {

	    /* Find distance between pointed and position counter. */	  
	    if (a > b){
	      factor = a - b;
	    }  else {
	      factor = b - a;
	    } /* end if (a > b) */

	    distance = distance + (running_total * factor);
	  
	    TEMP[b] = TEMP[b] + running_total;

	    if ( TEMP[b] > temp_mean) {  /* We overfilled the cell. */
	      running_total = TEMP[b] - temp_mean;
	      distance = distance - ((TEMP[b] - temp_mean) * factor);
	      TEMP[b] = temp_mean;
	    } else {
	      running_total = 0;
             
	    }  /* end if ( TEMP[b] > temp_mean)          */
	  }    /* end if ( TEMP[b] < temp_mean )         */
	}      /* end for (b = 0; b < num_stations; b++) */
      }        /* end if (running_total > 0)             */
    }          /* end if (TEMP[a] > temp_mean)           */
  }            /* end for (a = 0; a < num_stations; a++) */
  return distance;
}
  

/* =====================================================================
   This function receives a pointer to two arrays. First it fills a new
   array with the stored data in the original array. Next it generates
   a random SHUFFLE on that array and returns it.
   ===================================================================== */
int shuffle_array(float *ORIGINAL, float *NEW){

  int rand_num            = 0;
  int i, j                = 0;
  float temp              = 0;

  /* Copy the data from stored ORIGINAL array into our temporary NEW array */
  for (j = 0; j < MAX_ARRAY_SIZE; j++){NEW[j]=ORIGINAL[j];} 
  
  srand( (unsigned)time( NULL )*seed ); 

  for (i = 0; i < num_stations; i++){                
     
      rand_num = rand()%num_stations;                   /* Generate Random Number */

      temp = NEW[i]; NEW[i] = NEW[rand_num]; NEW[rand_num] = temp;     /* Shuffle */    
  }
  seed++;     /* Global int to remember place in sequence - avoids repeat numbers */
  return 0;
}
